@@ -27,6 +27,7 @@
 #include "JobTests.h"
 
 #include <cstdlib>
+#include <algorithm>
 
 #include <QtCore/QMutex>
 #include <QtTest/QtTest>
@@ -48,37 +49,13 @@
 
 #include "AppendCharacterJob.h"
 #include "AppendCharacterAndVerifyJob.h"
+#include "WaitForIdleAndFinished.h"
+#include "JobLoggingWeaver.h"
 
 QMutex s_GlobalMutex;
 
 using namespace ThreadWeaver;
 
-//Ensure that after the object is created, the weaver is idle and resumed.
-//Upon destruction, ensure the weaver is idle and suspended.
-class WaitForIdleAndFinished
-{
-public:
-    explicit WaitForIdleAndFinished(Queue *weaver)
-        : weaver_(weaver)
-    {
-        Q_ASSERT(weaver);
-        weaver_->finish();
-        Q_ASSERT(weaver_->isIdle());
-        weaver_->resume();
-    }
-
-    ~WaitForIdleAndFinished()
-    {
-        weaver_->resume();
-        weaver_->dequeue();
-        weaver_->finish();
-        weaver_->suspend();
-        Q_ASSERT(weaver_->isIdle());
-    }
-private:
-    Queue *weaver_;
-};
-
 void JobTests::initTestCase()
 {
     setDebugLevel(true, 1);
@@ -95,12 +72,41 @@ void JobTests::WeaverLazyThreadCreationTest()
     WaitForIdleAndFinished w(&weaver);
     Q_ASSERT(weaver.isIdle());
     QCOMPARE(weaver.currentNumberOfThreads(), 0);
+    weaver.setMaximumNumberOfThreads(1);
     weaver.stream() << new AppendCharacterJob(QChar('a'), &sequence);
     weaver.finish();
     QCOMPARE(weaver.currentNumberOfThreads(), 1);
     Q_ASSERT(weaver.isIdle());
 }
 
+void JobTests::ReduceWorkerCountTest()
+{
+    Queue weaver;
+    QString sequence;
+
+    WaitForIdleAndFinished w(&weaver);
+    Q_ASSERT(weaver.isIdle());
+    QCOMPARE(weaver.currentNumberOfThreads(), 0);
+    weaver.setMaximumNumberOfThreads(8);
+    weaver.stream() << new AppendCharacterJob(QChar('a'), &sequence);
+    weaver.finish();
+    QVERIFY(weaver.currentNumberOfThreads() >= 1);
+    weaver.setMaximumNumberOfThreads(1);
+    weaver.stream() << new AppendCharacterJob(QChar('b'), &sequence);
+    weaver.finish();
+    QCOMPARE(weaver.currentNumberOfThreads(), 1);
+    weaver.setMaximumNumberOfThreads(0);
+    weaver.stream() << new AppendCharacterJob(QChar('c'), &sequence);
+    weaver.reschedule();
+    // Unfortunately, there is no way to enforce that the last thread has exited:
+    // QCOMPARE(weaver.currentNumberOfThreads(), 0);
+    weaver.setMaximumNumberOfThreads(1);
+    weaver.stream() << new AppendCharacterJob(QChar('d'), &sequence);
+    weaver.finish();
+    QCOMPARE(weaver.currentNumberOfThreads(), 1);
+    QCOMPARE(sequence, QLatin1String("abcd"));
+}
+
 void JobTests::SimpleJobTest()
 {
     QString sequence;
@@ -154,7 +160,7 @@ void JobTests::CollectionQueueingTest()
     WaitForIdleAndFinished w(&weaver);
     weaver.suspend();
     weaver.stream() << jobCollection;
-    QCOMPARE(weaver.queueLength(), 1); //collection queues itself, and it's elements upon execution of self
+    QCOMPARE(weaver.queueLength(), 1); //collection queues itself, and its elements upon execution of self
     weaver.resume();
     weaver.finish();
     QCOMPARE(output.length(), 3);
@@ -171,11 +177,9 @@ class GeneratingCollection : public Collection
 {
 public:
     void run(JobPointer, Thread *) Q_DECL_OVERRIDE {
-        std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(),
-        [this](QChar it)
-        {
+        std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(), [this](QChar it) {
             *this << new AppendCharacterJob(it, &sequence_);
-        });
+        } );
     }
     QString sequence_;
 };
@@ -184,11 +188,9 @@ class GeneratingSequence : public Sequence
 {
 public:
     void run(JobPointer, Thread *) Q_DECL_OVERRIDE {
-        std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(),
-        [this](QChar it)
-        {
+        std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(), [this](QChar it) {
             *this << new AppendCharacterJob(it, &sequence_);
-        });
+        } );
     }
     QString sequence_;
 };
@@ -201,13 +203,35 @@ void JobTests::GeneratingCollectionTest()
 
     GeneratingCollection collection;
     WaitForIdleAndFinished w(Queue::instance());
-    stream() << make_job_raw(&collection);
+    stream() << collection;
     Queue::instance()->finish();
     QCOMPARE(collection.sequence_.count(), SequenceTemplate.length());
 }
 
 void JobTests::ShortJobSequenceTest()
 {
+    QString sequence;
+    Sequence jobSequence;
+    jobSequence << new AppendCharacterJob(QChar('a'), &sequence)
+                << new AppendCharacterJob(QChar('b'), &sequence)
+                << new AppendCharacterJob(QChar('c'), &sequence);
+
+    WaitForIdleAndFinished w(Queue::instance());
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    stream() << jobSequence;
+
+    Queue::instance()->finish();
+    QCOMPARE(sequence, QLatin1String("abc"));
+    QVERIFY(Queue::instance()->isIdle());
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+}
+
+void JobTests::ShortDecoratedJobSequenceTest()
+{
+    using namespace ThreadWeaver;
+    auto logger = new JobLoggingWeaver();
+    Queue queue(logger);
+
     QString sequence;
     JobPointer jobA(new AppendCharacterJob(QChar('a'), &sequence));
     JobPointer jobB(new AppendCharacterJob(QChar('b'), &sequence));
@@ -217,12 +241,14 @@ void JobTests::ShortJobSequenceTest()
     jobSequence->addJob(jobB);
     jobSequence->addJob(jobC);
 
-    WaitForIdleAndFinished w(Queue::instance());
-    Queue::instance()->enqueue(jobSequence);
+    WaitForIdleAndFinished w(&queue); Q_UNUSED(w);
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    queue.enqueue(jobSequence);
     // Job::DumpJobDependencies();
-    Queue::instance()->finish();
+    queue.finish();
     QCOMPARE(sequence, QLatin1String("abc"));
-    QVERIFY(Queue::instance()->isIdle());
+    QVERIFY(queue.isIdle());
+    QVERIFY(DependencyPolicy::instance().isEmpty());
 }
 
 void JobTests::EmptyJobSequenceTest()
@@ -313,7 +339,7 @@ void JobTests::EmitStartedOnFirstElementTest()
             break;
         }
         QTest::qWait(1);
-        debug(2, "JobTests::EmitStartedOnFirstElementTest: waiting (%i)\n", i);
+        TWDEBUG(2, "JobTests::EmitStartedOnFirstElementTest: waiting (%i)\n", i);
         qApp->processEvents();
     }
     QCOMPARE(collectionStartedSignalSpy.count(), 1);
@@ -693,20 +719,19 @@ void JobTests::MassiveJobSequenceTest()
 
 void JobTests::SimpleRecursiveSequencesTest()
 {
+    using namespace ThreadWeaver;
     QString sequence;
-    AppendCharacterJob jobB(QChar('b'), &sequence);
-    Sequence jobSequence1;
-    jobSequence1 << jobB;
 
-    AppendCharacterJob jobC(QChar('c'), &sequence);
-    AppendCharacterJob jobA(QChar('a'), &sequence);
+    Sequence jobSequence1;
+    jobSequence1 << new AppendCharacterJob(QChar('b'), &sequence);
+
     Sequence jobSequence2;
-    jobSequence2 << jobA;
-    jobSequence2 << jobSequence1;
-    jobSequence2 << jobC;
+    jobSequence2 << new AppendCharacterJob(QChar('a'), &sequence)
+                 << jobSequence1
+                 << new AppendCharacterJob(QChar('c'), &sequence);
 
     WaitForIdleAndFinished w(Queue::instance());
-    enqueue_raw(&jobSequence2);
+    stream() << jobSequence2;
     Queue::instance()->finish();
     QCOMPARE(sequence, QString("abc"));
 }
@@ -775,13 +800,10 @@ void JobTests::ResourceRestrictionPolicyBasicsTest()
     AppendCharacterJob g('g', &sequence);
     Collection collection;
     collection << a << b << c << d << e << f << g;
-    a.assignQueuePolicy(&restriction);
-    b.assignQueuePolicy(&restriction);
-    c.assignQueuePolicy(&restriction);
-    d.assignQueuePolicy(&restriction);
-    e.assignQueuePolicy(&restriction);
-    f.assignQueuePolicy(&restriction);
-    g.assignQueuePolicy(&restriction);
+    Q_FOREACH(AppendCharacterJob* job, QList<AppendCharacterJob*>() << &a << &b << &c << &d << &e << &f << &g) {
+        QMutexLocker l(job->mutex());
+        job->assignQueuePolicy(&restriction);
+    }
 
     WaitForIdleAndFinished w(Queue::instance());
     stream() << collection;
@@ -791,13 +813,11 @@ void JobTests::ResourceRestrictionPolicyBasicsTest()
 
 void JobTests::jobStarted(JobPointer)
 {
-    // qDebug() << "jobStarted";
     QVERIFY(thread() == QThread::currentThread());
 }
 
 void JobTests::jobDone(JobPointer)
 {
-    // qDebug() << "jobDone";
     QVERIFY(thread() == QThread::currentThread());
 }
 
@@ -934,7 +954,7 @@ void JobTests::IdDecoratorSingleAllocationTest()
 struct InstanceCountedJob : public Job {
     static QAtomicInt counter;
 
-    void run(JobPointer, Thread *)
+    void run(JobPointer, Thread *) Q_DECL_OVERRIDE
     {
     }
 
@@ -1043,7 +1063,7 @@ void JobTests::JobManualExitStatusTest()
     QCOMPARE(successful.status(), Job::Status_Success);
 }
 
-void JobTests::QueueStreamLifecycletest()
+void JobTests::QueueStreamLifecycleTest()
 {
     QString sequence;
     using namespace ThreadWeaver;
@@ -1055,5 +1075,138 @@ void JobTests::QueueStreamLifecycletest()
     QCOMPARE(sequence.count(), 3);
 }
 
+class SynchronizedNumbers {
+public:
+    void append(int number) {
+        QMutexLocker l(&mutex_);
+        numbers_.append(number);
+    }
+
+    bool isSorted() const {
+        QMutexLocker l(&mutex_);
+        return std::is_sorted(numbers_.cbegin(), numbers_.cend());
+    }
+
+    void sortChunks(int chunkSize) {
+        QMutexLocker l(&mutex_);
+        Q_ASSERT(numbers_.count() % chunkSize == 0);
+        auto start = numbers_.begin();
+        while(start!=numbers_.end()) {
+            auto stop = start;
+            std::advance(stop, chunkSize);
+            std::sort(start, stop);
+            start = stop;
+        }
+    }
+
+private:
+    QVector<int> numbers_;
+    mutable QMutex mutex_;
+};
+
+class GeneratingEnumeratorSequence : public ThreadWeaver::Sequence {
+public:
+    GeneratingEnumeratorSequence(SynchronizedNumbers* numbers, int start, int count)
+        : start_(start), count_(count), numbers_(numbers)
+    {}
+
+    void run(JobPointer, Thread*) Q_DECL_OVERRIDE {
+        numbers_->append(start_);
+        for(int index = start_ + 1; index < start_+count_; ++index) {
+            *this << new GeneratingEnumeratorSequence(numbers_, index, 1);
+        }
+    }
+
+private:
+    const int start_;
+    const int count_;
+    SynchronizedNumbers* numbers_;
+};
+
+class GeneratingEnumeratorCollection : public ThreadWeaver::Collection {
+public:
+    GeneratingEnumeratorCollection(SynchronizedNumbers* numbers, int start, int count)
+        : start_(start), count_(count), numbers_(numbers)
+    {}
+
+    void run(JobPointer, Thread*) Q_DECL_OVERRIDE {
+        numbers_->append(start_);
+        QVector<GeneratingEnumeratorCollection*> elements;
+        for(int index = start_ + 1; index < start_+count_; ++index) {
+            elements.append(new GeneratingEnumeratorCollection(numbers_, index, 1));
+        }
+        std::random_shuffle(elements.begin(), elements.end());
+        std::for_each(elements.begin(), elements.end(), [this](QVector<GeneratingEnumeratorCollection>::iterator it) {
+            *this << *it;
+        } );
+    }
+
+private:
+    const int start_;
+    const int count_;
+    SynchronizedNumbers* numbers_;
+};
+
+void JobTests::NestedGeneratingCollectionsTest()
+{
+    using namespace ThreadWeaver;
+    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+
+    SynchronizedNumbers numbers;
+    const int NumberOfSequences = 100;
+    const int ElementsPerSequence = 20;
+    Sequence sequence;
+    for(int index = 0; index < NumberOfSequences; ++index) {
+        sequence << new GeneratingEnumeratorCollection(&numbers, index * ElementsPerSequence, ElementsPerSequence);
+    }
+    stream() << sequence;
+    Queue::instance()->finish();
+    numbers.sortChunks(ElementsPerSequence);
+    QVERIFY(numbers.isSorted());
+}
+
+void JobTests::NestedGeneratingSequencesTest() {
+    using namespace ThreadWeaver;
+    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+
+    SynchronizedNumbers numbers;
+    const int NumberOfSequences = 100;
+    const int ElementsPerSequence = 20;
+    Sequence sequence;
+    for(int index = 0; index < NumberOfSequences; ++index) {
+        sequence << new GeneratingEnumeratorSequence(&numbers, index * ElementsPerSequence, ElementsPerSequence);
+    }
+    stream() << sequence;
+    Queue::instance()->finish();
+
+    QVERIFY(numbers.isSorted());
+}
+
+void JobTests::DeeperNestedGeneratingCollectionsTest()
+{
+    using namespace ThreadWeaver;
+    auto logger = new JobLoggingWeaver();
+    Queue queue(logger);
+    WaitForIdleAndFinished w(&queue); Q_UNUSED(w);
+    const int ElementsPerCollection = 20;
+    const int NumberOfBlocks = 2;
+    const int CollectionsPerBlock = 2;
+    SynchronizedNumbers numbers; Sequence sequence;
+    for(int block=0; block < NumberOfBlocks; ++block) {
+        auto col = new Collection();
+        for(int collection = 0; collection < CollectionsPerBlock; ++collection) {
+            const int start = (block * NumberOfBlocks + collection) * ElementsPerCollection;
+            *col << new GeneratingEnumeratorCollection(&numbers, start, ElementsPerCollection);
+        }
+        sequence << col;
+    }
+    queue.stream() << sequence;
+    queue.finish();
+    numbers.sortChunks(NumberOfBlocks*ElementsPerCollection);
+    QVERIFY(numbers.isSorted());
+}
+
+
+
 QTEST_MAIN(JobTests)
 
